/*
 * Copyright (C) 2020 CERN (www.cern.ch)
 *
 * Author: Miguel Jimenez Lopez <miguel.jimenez@sevensols.com>
 *
 * Released to the public domain as sample code to be customized.
 *
 * This work is part of the White Rabbit project, a research effort led
 * by CERN, the European Institute for Nuclear Research.
 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>

#include "stats-engine.h"
#include "fmc-dio-device.h"
#include "printf-log.h"
#include "file-log.h"
#include "irq-demo-private.h"

int main(int argc, char *argv[])
{
	fmc_dio_device dev;
	log_device gen_log;
	log_device stats_log;
	user_args user_arguments;
	log_device logs[2];
	unsigned int n_logs = 2;
	unsigned int user_stop = 0;
	stats_engine sengine;
	int ret = 0;

	gen_log = create_printf_log_device();
	stats_log = create_file_log_device(STATS_LOG_PATH);
	logs[0] = gen_log;
	logs[1] = stats_log;

	user_arguments = create_user_arguments();
	if(parse_user_arguments(argc, argv, user_arguments)) {
		show_help();
		ret = 1;
		goto out_log;
	}

	sengine = create_stats_engine();
	if(check_stats_engine(sengine)) {
		ret = 1;
		goto out_user_args;
	}
	attach_log_devices_to_stats_engine(sengine, logs, n_logs);
	enable_log_for_stats_engine(sengine);

	dev = create_fmc_dio_device(user_arguments->fmc_dev_path);
	if(open_fmc_dio_device(dev)) {
		ret = 1;
		goto out_stats_engine;
	}
	attach_log_devices_to_fmc_dio_device(dev, logs, n_logs);
	enable_log_for_fmc_dio_device(dev);

	disable_fmc_dio_device_all_irq(dev);
	enable_fmc_dio_device_irq(dev, TEST_FMC_DIO_CH);
	setup_fmc_dio_device_irq(dev, user_arguments->irq_period, TEST_FMC_DIO_COUNT);

	while(!user_stop) {
		demo_irq_process_loop(dev, sengine);
		user_stop = check_user_stop();
	}

	disable_fmc_dio_device_all_irq(dev);

out_fmc_dio_device:
	close_fmc_dio_device(dev);
out_stats_engine:
	destroy_stats_engine(sengine);
out_user_args:
	destroy_user_arguments(user_arguments);
out_log:
	destroy_log_device(stats_log);
	destroy_log_device(gen_log);

	return ret;
}

static void demo_irq_process_loop(fmc_dio_device fmc_dev, stats_engine engine)
{
	struct timespec *ts;
	unsigned int nts;
	while(!get_utc_ts_from_fmc_dio_device(fmc_dev, TEST_FMC_DIO_CH, &ts, &nts)) {
		process_timestamps_to_engine(engine, ts, nts);
		usleep(PROCESS_SLEEP_US);
	}
}

static void process_timestamps_to_engine(stats_engine engine, struct timespec *ts,
				unsigned int nts)
{
	for(int i = 0 ; i < nts ; i++) {
		add_usr_timestamp_to_stats_engine(engine, &ts[i]);
		run_stats_engine(engine);
	}
	free(ts);
}

static user_args create_user_arguments(void)
{
	user_args arguments;

	arguments = calloc(1, sizeof(*arguments));

	return arguments;
}

static int parse_user_arguments(int argc, char *argv[], user_args parsed_args)
{
	if(argc != 2 && argc != 3)
		return 1;

	parsed_args->fmc_dev_path = argv[1];

	if(argc == 3)
		sscanf(argv[2], "%d", &parsed_args->irq_period);
	else
		parsed_args->irq_period = TEST_FMC_DIO_PERIOD;

	return 0;
}

static void destroy_user_arguments(user_args args)
{
	if(args) {
		free(args);
	}
}

static void show_help(void)
{
	printf("usage: %s <FMC device path> [<irq period (ns)>]\n", PROG_NAME);
}

static unsigned int check_user_stop(void)
{
	int ret = 0;
	char c;
	fd_set fds;
	struct timeval tv;

	FD_ZERO(&fds);
	FD_SET(STDIN_FILENO, &fds);

	tv.tv_sec = 0;
	tv.tv_usec = 1;

	if(select(STDIN_FILENO+1, &fds, NULL, NULL, &tv) == 1) {
		c = getchar();
		if(c == 'q')
			ret = 1;
	}

	return ret;
}
